@@ -24,7 +24,7 @@ local t_concat = table.concat;
 local envloadfile = require"util.envload".envloadfile;
 local serialize = require "util.serialization".serialize;
 local lfs = require "lfs";
--- Extract directory seperator from package.config (an undocumented string that comes with lua)
+-- Extract directory separator from package.config (an undocumented string that comes with lua)
 local path_separator = assert ( package.config:match ( "^([^\n]+)" ) , "package.config not in standard form" )
 
 local prosody = prosody;
@@ -40,9 +40,10 @@ pcall(function()
 end);
 
 local _ENV = nil;
+-- luacheck: std none
 
 ---- utils -----
-local encode, decode;
+local encode, decode, store_encode;
 do
 	local urlcodes = setmetatable({}, { __index = function (t, k) t[k] = char(tonumber(k, 16)); return t[k]; end });
 
@@ -53,6 +54,12 @@ do
 	encode = function (s)
 		return s and (s:gsub("%W", function (c) return format("%%%02x", c:byte()); end));
 	end
+
+	-- Special encode function for store names, which historically were unencoded.
+	-- All currently known stores use a-z and underscore, so this one preserves underscores.
+	store_encode = function (s)
+		return s and (s:gsub("[^_%w]", function (c) return format("%%%02x", c:byte()); end));
+	end
 end
 
 if not atomic_append then
@@ -119,6 +126,7 @@ local function getpath(username, host, datastore, ext, create)
 	ext = ext or "dat";
 	host = (host and encode(host)) or "_global";
 	username = username and encode(username);
+	datastore = store_encode(datastore);
 	if username then
 		if create then mkdir(mkdir(mkdir(data_path).."/"..host).."/"..datastore); end
 		return format("%s/%s/%s/%s.%s", data_path, host, datastore, username, ext);
@@ -149,7 +157,8 @@ end
 
 local function atomic_store(filename, data)
 	local scratch = filename.."~";
-	local f, ok, msg, errno;
+	local f, ok, msg, errno; -- luacheck: ignore errno
+	-- TODO return util.error with code=errno?
 
 	f, msg, errno = io_open(scratch, "w");
 	if not f then
@@ -213,7 +222,7 @@ local function store(username, host, datastore, data)
 			os_remove(getpath(username, host, datastore));
 		end
 		-- we write data even when we are deleting because lua doesn't have a
-		-- platform independent way of checking for non-exisitng files
+		-- platform independent way of checking for nonexisting files
 	until ok;
 	return true;
 end
@@ -281,7 +290,7 @@ local function list_store(username, host, datastore, data)
 		os_remove(getpath(username, host, datastore, "list"));
 	end
 	-- we write data even when we are deleting because lua doesn't have a
-	-- platform independent way of checking for non-exisitng files
+	-- platform independent way of checking for nonexisting files
 	return true;
 end
 
@@ -311,8 +320,8 @@ local type_map = {
 }
 
 local function users(host, store, typ) -- luacheck: ignore 431/store
-	typ = type_map[typ or "keyval"];
-	local store_dir = format("%s/%s/%s", data_path, encode(host), store);
+	typ = "."..(type_map[typ or "keyval"] or typ);
+	local store_dir = format("%s/%s/%s", data_path, encode(host), store_encode(store));
 
 	local mode, err = lfs.attributes(store_dir, "mode");
 	if not mode then
@@ -321,9 +330,8 @@ local function users(host, store, typ) -- luacheck: ignore 431/store
 	local next, state = lfs.dir(store_dir); -- luacheck: ignore 431/next 431/state
 	return function(state) -- luacheck: ignore 431/state
 		for node in next, state do
-			local file, ext = node:match("^(.*)%.([dalist]+)$");
-			if file and ext == typ then
-				return decode(file);
+			if node:sub(-#typ, -1) == typ then
+				return decode(node:sub(1, -#typ-1));
 			end
 		end
 	end, state;
@@ -335,7 +343,7 @@ local function stores(username, host, typ)
 
 	local mode, err = lfs.attributes(store_dir, "mode");
 	if not mode then
-		return function() log("debug", err or (store_dir .. " does not exist")) end
+		return function() log("debug", "Could not iterate over stores in %s: %s", store_dir, err); end
 	end
 	local next, state = lfs.dir(store_dir); -- luacheck: ignore 431/next 431/state
 	return function(state) -- luacheck: ignore 431/state
